/************************************************************************/
/* netinfo_portable.c- Portability Source				*/
/************************************************************************/
/*
 *  (c)2008 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/

#include <fcntl.h>
#include <errno.h>

#include <netinfo/netinfo_portable.h>


/* BEGIN: w32 portable function declarations */
#ifdef WIN32

#include <stdint.h>
#include <winsock2.h>
#include <windows.h>


int pthread_attr_init (pthread_attr_t *attr)
{
	if (attr == NULL)
		return EINVAL;
	attr->h = NULL;
	return 0;
}


int pthread_attr_setdetachstate (pthread_attr_t *attr, int detachstate)
{
	if (attr == NULL)
		return EINVAL;
	return 0;
}


int pthread_mutex_init (pthread_mutex_t *mutex,
			pthread_mutexattr_t *mutexattr)
{
	if (mutex == NULL)
		return EINVAL;
	InitializeCriticalSection (&mutex->h);
	return 0;
}


int pthread_mutex_destroy (pthread_mutex_t *mutex)
{
	if (mutex == NULL)
		return EINVAL;
	DeleteCriticalSection (&mutex->h);
	return 0;
}


int pthread_mutex_lock (pthread_mutex_t *mutex)
{
	if (mutex == NULL)
		return EINVAL;
	EnterCriticalSection (&mutex->h);
	return 0;
}


int pthread_mutex_unlock (pthread_mutex_t *mutex)
{
	if (mutex == NULL)
		return EINVAL;
	LeaveCriticalSection (&mutex->h);
	return 0;
}


int  pthread_create(pthread_t  *thread,
			pthread_attr_t * attr,
		void * (*start_routine)(void *),
		void * arg)
{
	if (thread == NULL)
		return EINVAL;
	*thread = CreateThread(NULL,
				8192,
				(LPTHREAD_START_ROUTINE)start_routine,
				arg, 0, NULL);
	if (*thread == NULL)
		return GetLastError ();
	return 0;
}


int pthread_cond_init(pthread_cond_t *cond,
			pthread_condattr_t *cond_attr)
{
	if (cond == NULL)
		return EINVAL;
	// 2k8+: InitializeConditionVariable (&cond->h);
	cond->h = CreateEvent (NULL, FALSE, FALSE, NULL);
	if (cond->h == NULL)
		return GetLastError ();
	return 0;
}


int pthread_cond_destroy(pthread_cond_t *cond)
{
	if (cond == NULL)
		return EINVAL;
	// 2k8+: - No documentation on destroying a condition variable in win32 
	return 0;
}


int pthread_cond_signal (pthread_cond_t *cond)
{
	int		rc;
	if (cond == NULL || cond->h == NULL)
		return EINVAL;
	// 2k8+: WakeAllConditionVariable (&conf->h);
	rc = SetEvent (cond->h);
	if (rc == 0)
		return GetLastError ();
	return 0;
}


int pthread_cond_timedwait(pthread_cond_t *cond,
                        pthread_mutex_t *mutex,
                        const struct timespec *abstime)
{
	int		rc;
	uint32_t	ms;
	if (cond == NULL || mutex == NULL || abstime == NULL)
		return EINVAL;
	if (cond->h == NULL)
		return EINVAL;
	ms = (abstime->tv_sec * 1000) + (abstime->tv_nsec / 1000000);
	// 2k8+: rc = SleepConditionVariableCS (&cond->h, &mutex->h, ms);
	// Yes this is screwy... but hey, it's windows
	pthread_mutex_unlock (mutex);
	rc = WaitForSingleObject (cond->h, ms);
	pthread_mutex_lock (mutex);
	if (rc == 0) {
		rc = GetLastError();
		if (rc == WAIT_TIMEOUT)
			return ETIMEDOUT;
		if (rc == WSAECONNABORTED)
			return EPIPE;
		return rc;
	}
	return 0;
}


#ifndef localtime_r
struct tm *localtime_r (const time_t *time, struct tm *result)
{
	struct tm* tm1;
	if (time == NULL || result == NULL)
		return EINVAL;
	tm1 = localtime (time);
	memcpy (result, tm1, sizeof (struct tm));
	return 0;
}
#endif


int p_gethostname(char *buffer, int len)
{
	int	rc;
	int	l = len;
	rc = GetComputerNameEx(ComputerNamePhysicalDnsFullyQualified,
			buffer,
			&l);
	if (rc == 0)
		return -1;
	return 0;
}



#else // not win32

int p_gethostname(char *buffer, int len)
{
	int		rc;
	struct addrinfo hints, *info, *p;
	rc = gethostname(buffer, len);
	if (rc != 0)
		return rc;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_CANONNAME;
	rc = getaddrinfo(buffer, "http", &hints, &info);
	if (rc != 0)
		return rc;
	if (info == NULL)
		return rc;
	strncpy(buffer, info->ai_canonname, len);
	freeaddrinfo(info);
	return 0;
}


#endif


