/************************************************************************/
/* ni_inotify.c - Netinfo filesystem change monitoring plugin		*/
/************************************************************************/
/*
 *  Copyright 2005,2021 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/inotify.h>
#include <time.h>

#include <cutil/slist.h>
#include <cutil/strfuncs.h>

#include <netinfo/netinfo_portable.h>
#include <netinfo/netinfo.h>
#include <netinfo/nmodule.h>
#include <netinfo/conffile.h>

#include <fstraverse.h>
#include <fsobject.h>


#define	MODULE_NAME			"inotify"
#define	MODULE_VERSION			"0.1.0"

#ifndef WIN32
#define DEFAULT_CONFIGFILE		"passive.conf"
#define DIR_DELIMITER			"/"
#else
#define DEFAULT_CONFIGFILE		"passive.conf"
#define DIR_DELIMITER			"\\"
#define sleep(x)			Sleep (1000 * x)
#endif


#define BUFFER_SIZE			256
#define MAX_CONFIGMASK_ITEMS		16
#define	DEFAULT_HASH_WIDTH		1024

#define POLL_TIMEOUT_DEFAULT		5

// Module flags
#define MOD_F_RUNNING			1


// InotifyHandle flags
#define IN_F_RECURSIVE			1



typedef struct _handlehash_t {
	uint32_t	width;
	SList		**table;
} handlehash_t;


typedef struct	_InotifyHandle {
	int		id;			// inotify handle
	uint32_t	mask;			// inotify change event mask
	uint32_t	flags;			// entry flags (recursive, etc)
	char		*path;			// inotify path
	char		*obj_name;		// netinfo object name
	} InotifyHandle;


typedef struct _ModuleData {
	ModuleInfo	*moduleinfo;
	u_int32_t	flags;		// module state flags
	u_int32_t	poll_timeout;
	u_int32_t	curinterval;
	u_int32_t	hash_width;	// width of handle hash
	u_int32_t	curr_watches;	// current count of watches
	pthread_t	polltid;
	int		inotify_fd;	// inotify file descriptor
	SList		*handle_list;
	handlehash_t	handle_hash;	// hash of InotifyHandles indexed by watch descriptor (int)
	pthread_mutex_t loop_mutex;
	pthread_cond_t  loop_condition;
	EventChain	*event_chain;
        EventData       *event;
} ModuleData;


typedef struct _InotifyChangeMaskMap {
	char		*name;
	int		mask;
} InotifyChangeMaskMap;


/* Global Variables */
ModuleData	module_data;
InotifyChangeMaskMap change_mask_map[] = {
	{"ACCESS",		IN_ACCESS},
	{"MODIFY",		IN_MODIFY},
	{"ATTRIB",		IN_ATTRIB},
	{"CLOSE_WRITE",		IN_CLOSE_WRITE},
	{"CLOSE_NOWRITE",	IN_CLOSE_NOWRITE},
	{"OPEN",		IN_OPEN},
	{"MOVED_FROM",		IN_MOVED_FROM},
	{"MOVED_TO",		IN_MOVED_TO},
	{"CREATE",		IN_CREATE},
	{"DELETE",		IN_DELETE},
	{"DELETE_SELF",		IN_DELETE_SELF},
	{"MOVE_SELF",		IN_MOVE_SELF},
	{NULL,			0},
};


/* Function Prototypes */
void *poll_thread (void *arg);
int parse_configfile (char *filepath);
int inotify_path_add (char *path, char *objname, int mask, uint32_t flags);
int inotify_path_add_recursive(char *path, char *objname, int mask, uint32_t flags);


int handlehash_init(handlehash_t *hh, uint32_t width)
{
	if (hh == NULL || width < 32)
		return EINVAL;
	hh->width = width;
	hh->table = calloc(hh->width, sizeof(InotifyHandle*));
	if (!hh->table)
		return ENOMEM;
	return 0;
}

InotifyHandle *handlehash_get_by_id(handlehash_t *hh, int id)
{
	int		m = id % hh->width;
	SList		*l = hh->table[m];
	InotifyHandle	*h;
	while (l) {
		h = slist_get_data(l);
		if (h->id == id)
			return h;
		l = slist_get_next(l);
	}
	return NULL;
}


int handlehash_add(handlehash_t *hh, InotifyHandle* handle)
{
	int		rc;
	int		m = handle->id % hh->width;
	SList		*l = hh->table[m];
	InotifyHandle	*h;
	while (l) {
		h = slist_get_data(l);
		if (h->id == handle->id)
			return EEXIST;
		l = slist_get_next(l);
	}
	rc = slist_item_add(&hh->table[m], handle);
	return rc;
}


int handlehash_remove(handlehash_t *hh, InotifyHandle* handle)
{
	int		rc;
	int		m = handle->id % hh->width;
	SList		*l = hh->table[m];
	InotifyHandle	*h;
	while (l) {
		h = slist_get_data(l);
		if (h->id == handle->id) {
			rc = slist_item_remove(&hh->table[m], handle);
			return rc;
		}
		l = slist_get_next(l);
	}
	return ENOENT;
}


int handlehash_free(handlehash_t *hh)
{
	int		i;
	int		rc;
	SList		*l;
	InotifyHandle	*h;
	for (i=0 ; i<hh->width ; i++) {
		l = hh->table[i];
		if (!l)
			continue;
		while(i) {
			h = slist_get_data(l);
			free(h->obj_name);
			free(h->path);
			free(h);
			l = slist_get_next(l);
		}
		rc = slist_free(&hh->table[i]);
		if(rc)
			return rc;
	}
	free(hh->table);
	hh->width = 0;
	return 0;
}



int netinfo_module_init (ModuleInfo *mi)
{
	int rc;
	int flags;
	char buffer[BUFFER_SIZE];
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (&module_data, 0, sizeof (ModuleData));
	module_data.moduleinfo = mi;
	module_data.moduleinfo->name = MODULE_NAME;
	module_data.poll_timeout = POLL_TIMEOUT_DEFAULT;
	module_data.curinterval = 0;
	module_data.flags = MOD_F_RUNNING;
	module_data.hash_width = DEFAULT_HASH_WIDTH;
	rc = handlehash_init(&module_data.handle_hash, module_data.hash_width);
	vector_log_entry (module_data.moduleinfo,
		"Handle hash width set to %d\n", module_data.hash_width);

	rc = pthread_mutex_init (&module_data.loop_mutex, NULL);
	if (rc != 0) {
		vector_log_entry (mi,
				"Failed to initialize loopmutex rc=%d\n",
				rc);
		handlehash_free(&module_data.handle_hash);
		return NETINFO_RC_MODULEERROR;
	}
	rc = pthread_cond_init (&module_data.loop_condition, NULL);
	if (rc != 0) {
		vector_log_entry (mi,
				"Failed to initialize loopcondition rc=%d\n",
				rc);
		pthread_mutex_destroy (&module_data.loop_mutex);
		handlehash_free(&module_data.handle_hash);
		return NETINFO_RC_MODULEERROR;
	}

#ifndef inotify_init1
	module_data.inotify_fd = inotify_init();
#else
	module_data.inotify_fd = inotify_init1(IN_NONBLOCK);
#endif
	if (module_data.inotify_fd == -1) {
		vector_log_entry (mi,
				"Failed to initialize inotify rc=%d\n",
				errno);
		pthread_mutex_destroy (&module_data.loop_mutex);
		handlehash_free(&module_data.handle_hash);
		return NETINFO_RC_MODULEERROR;
	}
#ifndef inotify_init1
	flags = fcntl(module_data.inotify_fd, F_GETFL, 0);
	rc = fcntl(module_data.inotify_fd, F_SETFL, flags | O_NONBLOCK);
	if (rc == -1) {
		vector_log_entry (mi,
				"Failed to setup nonblocking fd for inotify rc=%d\n",
				errno);
		rc = close(module_data.inotify_fd);
		pthread_mutex_destroy (&module_data.loop_mutex);
		handlehash_free(&module_data.handle_hash);
		return NETINFO_RC_MODULEERROR;
	}
#endif

	rc = vector_eventchain_open (module_data.moduleinfo,
				&module_data.event_chain);

	rc = netinfo_get_component_filepath (buffer, BUFFER_SIZE, "etc", DEFAULT_CONFIGFILE);
	rc = parse_configfile (buffer);

	rc = pthread_create (&module_data.polltid,
				NULL,
				poll_thread,
				NULL);
	if (rc == 0) {
		vector_log_entry (mi,
				"Started poll thread (tid=%d) interval=%d\n",
				module_data.polltid,
				module_data.poll_timeout);
	} else {
		vector_log_entry (mi,
				"Failed to start poll thread.\n");
		rc = close(module_data.inotify_fd);
		pthread_mutex_destroy (&module_data.loop_mutex);
		handlehash_free(&module_data.handle_hash);
		return NETINFO_RC_MODULEERROR;
	}
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_close (ModuleInfo *mi)
{
	int		rc;
	SList		*litem = NULL;
	InotifyHandle	*h = NULL;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;

	module_data.flags &= ~(MOD_F_RUNNING);
	pthread_cond_signal (&module_data.loop_condition);
	// TODO: pthread_join();	// wait for worker thread to exit

	litem = module_data.handle_list;
	while (litem != NULL) {
		h = (InotifyHandle *)slist_get_data (litem);
		if (h != NULL) {
			rc = inotify_rm_watch(module_data.inotify_fd, h->id);
			if (h->obj_name)
				free(h->obj_name);
			if (h->path)
				free(h->path);
			free (h);
		}
		litem = slist_get_next (litem);
	}
	rc = slist_free (&module_data.handle_list);
	rc = close(module_data.inotify_fd);
	rc = vector_eventchain_close (&module_data.event_chain);
	(void)rc;
	return NETINFO_RC_SUCCESS;
}


int remove_watch_handle_by_id(int wd)
{
	int		rc;
	InotifyHandle	*h;
	if (wd < 0)
		return EINVAL;
	h = handlehash_get_by_id(&module_data.handle_hash, wd);
	if (!h) {
		return ENOENT;
	}
	rc = slist_item_remove(&module_data.handle_list, h);
	if (rc != ESUCCESS) {
		return rc;
	}
	rc = handlehash_remove(&module_data.handle_hash, h);
	free(h->obj_name);
	free(h->path);
	free(h);
	module_data.curr_watches--;
	return NETINFO_RC_SUCCESS;
}


int handle_inotify_events(void)
{
	char				data[BUFFER_SIZE];
	char				clist[BUFFER_SIZE];
	char				buf[4096]
					__attribute__ ((aligned(__alignof__(struct inotify_event))));
	const struct inotify_event	*event;
	int				rc;
	int				eventrc;
	int				dlen;
	ssize_t				len;
	char				*ptr;
	InotifyHandle			*h = NULL;

	while (1) {

		len = read(module_data.inotify_fd, buf, sizeof(buf));
		if (len == -1) {
			if (errno == EAGAIN)
				break; // no more data to read
			// some other error
			vector_log_entry (module_data.moduleinfo,
				"Error reading from inotify descriptor rc=%d.\n", errno);
			return errno;
		}

		for (ptr = buf; ptr < buf + len; ptr += sizeof(struct inotify_event) + event->len) {
			event = (const struct inotify_event *) ptr;

			h = handlehash_get_by_id(&module_data.handle_hash, event->wd);
			if (!h) {
				vector_log_entry (module_data.moduleinfo,
					"Watch event received for invalid descriptor id=%d.\n",
					event->wd);
				continue;
			}

			if (event->mask & IN_IGNORED) {
				// the watched object has been deleted
				// remove the watch
				vector_log_entry (module_data.moduleinfo,
					"Removing watch path=%s id=%d\n",
					h->path,
					event->wd);
				rc = remove_watch_handle_by_id(event->wd);
				continue;
			}

			clist[0] = 0;
			if (event->mask & IN_CREATE) {
				strcat(clist, "CREATE");
			}
			if (event->mask & IN_MODIFY) {
				if (clist[0] != 0)
					strcat(clist, ",");
				strcat(clist, "MODIFY");
			}
			if (event->mask & IN_ATTRIB) {
				if (clist[0] != 0)
					strcat(clist, ",");
				strcat(clist, "ATTRIBUTE");
			}
			if (event->mask & IN_MOVE) {
				if (clist[0] != 0)
					strcat(clist, ",");
				strcat(clist, "MOVED");
			}
			if (event->mask & IN_DELETE) {
				if (clist[0] != 0)
					strcat(clist, ",");
				strcat(clist, "DELETE");
			}
			if (event->mask & IN_ACCESS) {
				if (clist[0] != 0)
					strcat(clist, ",");
				strcat(clist, "ACCESS");
			}
			if (event->mask & IN_DELETE_SELF) {
				if (clist[0] != 0)
					strcat(clist, ",");
				strcat(clist, "DELETE_SELF");
			}
			if (event->mask & IN_MOVE_SELF) {
				if (clist[0] != 0)
					strcat(clist, ",");
				strcat(clist, "MOVE_SELF");
			}

			eventrc = 2;
			if (event->len) {
				dlen = snprintf (data,
					BUFFER_SIZE,
					"%s event(s) occurred on the object %s/%s\n",
					clist,
					h->path,
					event->name);
			} else {
				dlen = snprintf (data,
					BUFFER_SIZE,
					"%s event(s) occurred on the object %s\n",
					clist,
					h->path);
			}
			eventdata_clear (module_data.event);
			rc = eventdata_set (module_data.event,
					EVENT_FLAG_ONESHOT,
					eventrc,
					h->obj_name,
					data,
					dlen);
			rc = vector_eventchain_emit_event (module_data.event_chain,
							module_data.event);

			if (event->mask & IN_CREATE && event->mask & IN_ISDIR) {
				// add watch if required
				if (h->flags & IN_F_RECURSIVE) {
					dlen = snprintf (data,
						BUFFER_SIZE,
						"%s/%s",
						h->path,
						event->name);
					rc = inotify_path_add (data, h->obj_name, h->mask, h->flags);
					if (rc == -1) {
						vector_log_entry (module_data.moduleinfo,
							"Error adding watch path=%s rc=%d\n",
							data,
							rc);
					}
					if (h->flags & IN_F_RECURSIVE) {
						rc = inotify_path_add_recursive (data,
										h->obj_name,
										h->mask,
										h->flags);
						if (rc == -1) {
							vector_log_entry (module_data.moduleinfo,
								"Error recursing watch path=%s rc=%d\n",
								data,
								rc);
						}
					}
				}
			}

		}
	}
	(void)rc;
	return NETINFO_RC_SUCCESS;
}


void *poll_thread (void *arg)
{
	int		rc;
	int		prc;
	int		timeout;
//	struct timespec ts;
//	struct timeval  tv;
	nfds_t		nfds;
	struct pollfd	fds[2];
	ModuleData	*md = &module_data;

	sleep (2); /*FIXME:  module is loaded before netinfod can initialize SSL context */
	
        rc = eventdata_new (&module_data.event);
	(void)rc;
	eventdata_clear (module_data.event);

	// poll setup
	nfds = 1;
	fds[0].fd = module_data.inotify_fd;
	fds[0].events = POLLIN;
	timeout = POLL_TIMEOUT_DEFAULT * 1000;

	md->curinterval = 0;
	while (1) {

		// Wait for inotify events or timeout
		prc = poll(fds, nfds, timeout);
		if (prc == -1) {
			if (errno != EINTR)
				break;
		}

		if (prc > 0 ) {
			if (fds[0].revents & POLLIN) {
				rc = handle_inotify_events();
			}
		}


/* version2: using a mutex/cond
		rc = gettimeofday (&tv, NULL);
		ts.tv_sec = tv.tv_sec + md->poll_timeout;
		ts.tv_nsec = tv.tv_usec * 1000;
		pthread_mutex_lock (&md->loopmutex);
		rc = pthread_cond_timedwait (&md->loopcondition,
				&md->loopmutex,
				&ts);
		pthread_mutex_unlock (&md->loopmutex);
		if (rc != ETIMEDOUT)
			break;
*/

		// Handle global state events
		if (!(module_data.flags & MOD_F_RUNNING))
			break;


	}

	eventdata_free (&module_data.event);
	pthread_exit (NULL);
	return NULL;
}


int inotify_changemask_map_name(char *name)
{
	int		i;
	for (i=0 ; change_mask_map[i].name!=NULL ; i++) {
		if (!strcmp(name, change_mask_map[i].name)) {
			return change_mask_map[i].mask;
		}
	}
	return -1;
}



int inotify_path_add (char *path, char *objname, int mask, uint32_t flags)
{
	int		rc;
	InotifyHandle	*h = NULL;

	h = (InotifyHandle *)calloc (1, sizeof (InotifyHandle));
	if (!h)
		return -1;
	h->flags = flags;
	h->mask = mask;
	h->obj_name = strdup(objname);
	h->path = strdup (path);
	h->id = inotify_add_watch(module_data.inotify_fd,
				h->path,
				h->mask);
	if (h->id == -1) {
		rc = errno;
		vector_log_entry (module_data.moduleinfo,
				"Failed to setup watch for path='%s' mask=%04x errno=%d\n",
				h->path,
				h->mask,
				errno);
		free(h->obj_name);
		free(h->path);
		free(h);
		errno = rc;
		return -1;
	}
	rc = slist_item_add (&module_data.handle_list, h);
	handlehash_add(&module_data.handle_hash, h);
	module_data.curr_watches++;
	vector_log_entry (module_data.moduleinfo,
			"Watching path='%s' mask=%04x id=%d\n",
			path,
			mask,
			h->id);
	return NETINFO_RC_SUCCESS;
}


int inotify_path_add_recursive(char *path, char *objname, int mask, uint32_t flags)
{
	int		rc;
	FSTrav		*fst = NULL;
	FSObject	*fso = NULL;

	fst = fs_trav_open (path, FST_FLAG_RECURSIVE|FST_FLAG_SHOWHIDDEN);
	if (fst == NULL) {
		vector_log_entry (module_data.moduleinfo,
				"Failed to initiate traverse for path='%s', ignoring\n",
				path);
		return NETINFO_RC_INVALIDARG;
	}
	while (1) {
		rc = fs_trav_next_entry (fst, &fso);
		if (rc != 0) {
			break;
		}
		if (fso->object_type == OBJECT_TYPE_DIRECTORY) {
			rc = inotify_path_add (fso->object_path, objname, mask, flags);
			(void)fs_object_free(&fso);
			if (rc == -1) {
				// error message is logged in inotify_path_add()
				continue;
			}
		}
	}
	rc = fs_trav_close (fst);
	return NETINFO_RC_SUCCESS;
}


int parse_configfile (char *filepath)
{
	int			rc;
	int			i;
	int			cmlen;
	uint32_t		m;
	uint32_t		mask;
	uint32_t		flags;
	char			buffer[BUFFER_SIZE];
	char			*directive = NULL;
	char			*attribs = NULL;
	char			*attrib = NULL;
	char			*cmlist = NULL;
	char			**cmarray = NULL;
	char			*path = NULL;
	char			*objname = NULL;
	ConfigFile		*cf = NULL;

	if (filepath == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = configfile_open (&cf, filepath);
	if (rc != NETINFO_RC_SUCCESS)
		return rc;
	rc = configfile_search_begin (cf, "WatchChange");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		directive = strtok (buffer, "'\" \t");
		if (directive[0] == '#') {
			continue;
		}
		
		path = strtok (NULL, "'\"\t");
		flags = 0;
		objname = NULL;
		mask = 0;
		attribs = strtok (NULL, " \t\r\n");
		attrib = strtok (attribs, "= \t\r\n");
		while (attrib != NULL) {
			if (!strcmp (attrib, "objectname")) {
				objname = strtok (NULL, ", \t\r\n");
			} else if (!strcmp (attrib, "recursive")) {
				rc = atol (strtok (NULL, ", \t\r\n"));
				if (rc == 1) {
					flags |= IN_F_RECURSIVE;
				}
			} else if (!strcmp (attrib, "changemask")) {
				cmlist = strtok (NULL, ", \t\r\n");
				cmlen = MAX_CONFIGMASK_ITEMS;
				rc = strnsplit (cmlist,
						strlen (cmlist),
						'|',
						&cmarray,
						&cmlen);
				for (i=0 ; i< cmlen ; i++) {
					m = inotify_changemask_map_name(cmarray[i]);
					if (m == -1) {
						vector_log_entry (module_data.moduleinfo,
								"Invalid mask item '%s' in watch for path='%s', invalidating\n",
								cmarray[i],
								path);
						mask = 0;
						break;
					}
					mask |= m;
				}
				rc = strarray_free(&cmarray);
			}
			attrib = strtok (NULL, "= \t\r\n");
		}
		if (objname == NULL) {
			vector_log_entry (module_data.moduleinfo,
					"Objectname is required on watch for path='%s', ignoring\n",
					path);
			continue;
		}
		if (mask == 0) {
			vector_log_entry (module_data.moduleinfo,
					"Change mask cannot be zero on watch for path='%s', ignoring\n",
					path);
			continue;
		}
		// Add a watch for the specified path
		rc = inotify_path_add (path, objname, mask, flags);
		if (rc == -1) {
			// error message is logged in inotify_path_add()
			continue;
		}
		if (flags & IN_F_RECURSIVE) {
			// add watches recursively under the specified path
			rc = inotify_path_add_recursive(path, objname, mask, flags);
			if (rc == -1) {
				// error message is logged in inotify_path_add_recursive()
				continue;
			}
		}
	}
	rc = configfile_search_end (cf);
	rc = configfile_close (&cf);
	vector_log_entry (module_data.moduleinfo,
		"Successfully created %d inotify handles\n",
		module_data.curr_watches);

	return NETINFO_RC_SUCCESS;
}


